#!/usr/bin/env python
# coding: utf-8

from abc import ABC, abstractmethod

from napoleontoolbox.file_saver import dropbox_file_saver

from napoleontoolbox.forecasting import forecasting_model
from sklearn.metrics import mean_squared_error
from sklearn.metrics import accuracy_score
from sklearn.metrics import confusion_matrix

from napoleontoolbox.signal import signal_utility
from napoleontoolbox.rebalancing import time_series_forecasting
from napoleontoolbox.utility import metrics
import datetime

import json

import pandas as pd
import numpy as np


class AbstractRunner(ABC):
    def __init__(self, starting_date = None, running_date = None, drop_token=None, dropbox_backup = True, X_y_pkl_file_name_suffix ='_X_y.pkl', local_root_directory='../data/', user = 'napoleon'):
        super().__init__()
        self.X_y_pkl_file_name_suffix=X_y_pkl_file_name_suffix
        self.local_root_directory=local_root_directory
        self.user=user
        self.dropbox_backup = dropbox_backup
        self.dbx = dropbox_file_saver.NaPoleonDropboxConnector(drop_token=drop_token,dropbox_backup=dropbox_backup)
        self.running_date = running_date
        self.starting_date = starting_date
        self.starting_days = 10



    @abstractmethod
    def runTrial(self,saver,  seed,  n, s, s_eval, n_fft, n_wavelet, add_wavelet_detail, calibration_step,  optimize_threshold, signal_type, idios_string):
        pass

class ForecasterRunner(AbstractRunner):
    def runTrial(self, saver, seed, n, s, s_eval, n_fft, n_wavelet, add_wavelet_detail, calibration_step, optimize_threshold, signal_type, idios_string):
        if n == 0:
            n = None
        if s_eval == 0:
            s_eval = None
        idios = json.loads(idios_string)
        common_params ={
            'n':n,
            's':s,
            's_eval':s_eval,
            'n_fft':n_fft,
            'n_wavelet':n_wavelet,
            'add_wavelet_detail':add_wavelet_detail,
            'calibration_step':calibration_step,
            'optimize_threshold':optimize_threshold,
            'signal_type' : signal_type
        }
        common_params.update(idios)
        saving_key = json.dumps(common_params, sort_keys=True)
        saving_key = signal_utility.convert_to_sql_column_format(saving_key)
        check_run_existence = saver.checkRunExistence(saving_key)
        if check_run_existence:
            return
        from napoleontoolbox.connector import pickle_utility
        #raw_data = pickle_utility.read_pickle(self.local_root_directory+str(n_fft)+'_'+ str(n_wavelet)+'_'+str(add_wavelet_detail)+self.X_y_pkl_file_name_suffixself.local_root_directory+str(n_fft)+'_'+ str(n_wavelet)+'_'+str(add_wavelet_detail)+self.X_y_pkl_file_name_suffix)
        raw_data = pd.read_pickle(self.local_root_directory+str(n_fft)+'_'+ str(n_wavelet)+'_'+str(add_wavelet_detail)+self.X_y_pkl_file_name_suffix)
        print('size before filtering')
        print(raw_data.shape)

        raw_data=raw_data[raw_data.index >= (self.starting_date+datetime.timedelta(days=max(n_fft, n_wavelet)+self.starting_days))]
        raw_data=raw_data[raw_data.index <= self.running_date]
        print('size after filtering')
        print(raw_data.shape)
        X, y = raw_data.loc[:, raw_data.columns != 'target'], raw_data['target']
        print('predictors shape')
        print(X.shape)
        print('output shape')
        print(y.shape)
        chosen_method = signal_type
        model = forecasting_model.ForecasterWrapper(method=chosen_method)

        # Compute rolling weights
        forecasted_series, features_importances = time_series_forecasting.rolling_forecasting(
            model,
            X,
            y,
            n=n,
            s=s,
            s_eval = s_eval,
            calibration_step = calibration_step,
            optimize_threshold = optimize_threshold,
            method = chosen_method,
            display = True)

        print('forecasting done')
        features_importances = features_importances.sum(axis = 0)
        print('features importance')
        print(features_importances.sum())

        forecasted_series[np.isnan(forecasted_series)] = 0
        forecasted_series[np.isinf(forecasted_series)] = 0
        matrix = confusion_matrix(y > 0, forecasted_series > 0)
        rmse = mean_squared_error(y, forecasted_series)
        accuracy = accuracy_score(y > 0, forecasted_series > 0)


        print('rmse '  + str(rmse))
        print('accuracy')
        print(accuracy)
        print('confusion matrix')
        print(matrix)

        # print(forecasted_series.shape)
        # if not filter_best_predictors:
        #     features_importances.to_pickle(local_root_directory + 'importances.pkl')
        #
        # perf_df = signal_utility.reconstitute_prediction_perf(y_pred=forecasted_series>0,y_true = y, transaction_cost=False,
        #                                                     print_turnover=False)
        #
        # perf_df = perf_df[perf_df.index >= '2002-02-01']
        # perf_df.to_pickle(local_root_directory +'perf_df.pkl')
        # #perf_df = pd.read_pickle(local_root_directory +'perf_df.pkl')
        # perf_df[['reconstituted_perf','reconstituted_close']].plot()


        perf_df = signal_utility.reconstitute_prediction_perf(y_pred=forecasted_series>0, y_true = y, transaction_cost=False, print_turnover=False)
        sharpe_strat = metrics.sharpe(perf_df['perf_return'].dropna(), period= 252, from_ret=True)
        sharpe_under = metrics.sharpe(perf_df['close_return'].dropna(), period= 252, from_ret=True)
        print('underlying sharpe')
        print(sharpe_under)
        print('strat sharpe')
        print(sharpe_strat)
        saver.saveAll(saving_key, perf_df)





import lightgbm as lgb

import numpy as np

from scipy.stats import uniform, randint

from sklearn.datasets import load_breast_cancer, load_diabetes, load_wine
from sklearn.metrics import auc, accuracy_score, confusion_matrix, mean_squared_error
from sklearn.model_selection import cross_val_score, GridSearchCV, KFold, RandomizedSearchCV, train_test_split
from sklearn.linear_model import Lasso
import xgboost as xgb
from sklearn import datasets, linear_model

def report_best_scores(results, n_top=3):
    for i in range(1, n_top + 1):
        candidates = np.flatnonzero(results['rank_test_score'] == i)
        for candidate in candidates:
            print("Model with rank: {0}".format(i))
            print("Mean validation score: {0:.3f} (std: {1:.3f})".format(
                  results['mean_test_score'][candidate],
                  results['std_test_score'][candidate]))
            print("Parameters: {0}".format(results['params'][candidate]))
            print("")

class ForecasterWrapper:
    """ Base object to roll a neural network model.
    Rolling over a time axis with a train period from `t - n` to `t` and a
    testing period from `t` to `t + s`.
    Parameters
    ----------
    X, y : array_like
        Respectively input and output data.
    f : callable, optional
        Function to transform target, e.g. ``torch.sign`` function.
    index : array_like, optional
        Time index of data.
    Methods
    -------
    __call__
    Attributes
    ----------
    n, s, r : int
        Respectively size of training, testing and rolling period.
    b, e, T : int
        Respectively batch size, number of repass_steps and size of entire dataset.
    t : int
        The current time period.
    y_eval, y_test : np.ndarray[ndim=1 or 2, dtype=np.float64]
        Respectively evaluating (or training) and testing predictions.
    """
    def __init__(self, method = 'standard'):
        """ Initialize shape of target. """
        if method == 'standard':
            self.model =  linear_model.LinearRegression()
        if method == 'lasso':
            alpha = 0.1
            self.model = Lasso(alpha=alpha, fit_intercept=False, max_iter=5000)
        if method == 'lgbm':
            self.model = lgb.LGBMRegressor()
        if method == 'xgb':
            self.model = xgb.XGBRegressor()

    def calibrate(self, X, y, method = 'standard'):
        if method == 'standard':
            #nothing to calibrate here
            self.model = linear_model.LinearRegression()
        if method == 'lasso':
            param_grid = {
                'alpha': [0.01, 0.1,0.5,1],
            }
            gbm_grid = GridSearchCV(self.model, param_grid, cv=2)
            gbm_grid.fit(X, y)
            print('Best parameters found by grid search are:', gbm_grid.best_params_)
            best_params = gbm_grid.best_params_
            self.model = Lasso(alpha=best_params['alpha'], fit_intercept=False, max_iter=5000)
        if method == 'lgbm':
            param_grid = {
                'boosting_type': ['gbdt'],
                'num_leaves': [int(x) for x in np.linspace(start=10, stop=30, num=10)],
                'max_depth': [int(x) for x in np.linspace(start=-1, stop=7, num=3)],
                'learning_rate': [0.001,0.01,0.2],
                'n_estimators': [int(x) for x in np.linspace(start=40, stop=120, num=40)]
            }
            gbm_grid = GridSearchCV(self.model, param_grid, cv=2)
            gbm_grid.fit(X, y)
            print('Best parameters found by grid search are:', gbm_grid.best_params_)
            best_params = gbm_grid.best_params_
            self.model.set_params(**best_params)
            return best_params
        if method == 'xgb':
            params = {
                "colsample_bytree": uniform(0.7, 0.3),
                "gamma": uniform(0, 0.5),
                "learning_rate": uniform(0.03, 0.3),  # default 0.1
                "max_depth": randint(2, 6),  # default 3
                "n_estimators": randint(100, 150),  # default 100
                "subsample": uniform(0.6, 0.4)
            }
            cv = 2
            n_iter = 50
            # cv=3
            # n_iter = 200
            search = RandomizedSearchCV(self.model, param_distributions=params, random_state=42, n_iter=n_iter, cv=cv,
                                        verbose=1, n_jobs=1, return_train_score=True)
            search.fit(X, y)
            print('reporting best scores')
            report_best_scores(search.cv_results_, 1)

    def fit(self, X_train, y_train, X_val, y_val, method = 'standard'):
        if method == 'standard':
            self.model.fit(X_train, y_train)
        if method == 'lasso':
            self.model.fit(X_train, y_train)
        if method == 'lgbm':
            self.model.fit(X_train, y_train,eval_set=[(X_val, y_val)],eval_metric= ['l1','l2'],early_stopping_rounds=5,verbose= 0)
        if method == 'xgb':
            self.model.fit(X_train, y_train, eval_set=[(X_val, y_val)], eval_metric=['rmse', 'logloss'],
                           early_stopping_rounds=5, verbose=0)

    def predict(self, X_test, method = 'standard'):
        y_pred = self.model.predict(X_test)
        return y_pred

    def get_features_importance(self, features_names, method = 'standard'):
        run_importances = {}
        if method == 'lgbm':
            for (name, imp) in zip(features_names, self.model.feature_importances_):
                run_importances[name] = imp
        if method == 'standard':
            for (name, imp) in zip(features_names, self.model.coef_):
                run_importances[name] = imp
        return run_importances
